#include "core/Window.h"

#include <GLFW/glfw3.h>
#include <core/HashMap.h>
#include <core/Logger.h>
#include <core/Utility.h>
#include <stdio.h>

static bool dummyWindowRunHandler(void*) {
    return !coreWindowShouldClose();
}

static void dummyWindowTickHandler(void*) {
}

static void dummyWindowRenderHandler(void*, float) {
}

static GLFWwindow* window = nullptr;
static CoreIntVector2 size = {0};
static bool sizeChanged = false;
static CoreWindowRunHandler runHandler = dummyWindowRunHandler;
static CoreWindowTickHandler tickHandler = dummyWindowTickHandler;
static CoreWindowRenderHandler renderHandler = dummyWindowRenderHandler;
static void* runHandlerData = nullptr;
static void* tickHandlerData = nullptr;
static void* renderHandlerData = nullptr;
static i64 nanosPerTick = 1;

#define CLOCK_SIZE 16
typedef struct {
    i64 last;
    i64 sum;
    i64 values[CLOCK_SIZE];
    size_t index;
} Clock;
static Clock fps = {0};
static Clock tps = {0};

// static List<uint32> input;
// static int inputCursor = 0;
// static int inputLimit = 256;
// static bool inputActive = false;

typedef struct {
    char name[31];
    bool released;
    int key;
    int downTime;
    int upEvents;
    int downEvents;
    int controllerUp;
    int controllerDown;
} ButtonData;
#define BUTTONS 100
static ButtonData buttons[BUTTONS] = {{.name = "unknown"}};
static size_t buttonIndex = 1;
#define KEYS (GLFW_KEY_LAST + 1)
static CoreButton keyToButton[KEYS] = {0};
#define GAMEPAD_BUTTONS (GLFW_GAMEPAD_BUTTON_LAST + 1)
static CoreButton gamepadToButton[GAMEPAD_BUTTONS] = {0};
#define MOUSE_BUTTONS (GLFW_MOUSE_BUTTON_LAST + 1)
static CoreButton mouseToButton[MOUSE_BUTTONS] = {0};
static CoreVector2 lastMousePosition = {0};
static CoreVector2 mousePosition = {0};
static int activeController = -1;
static GLFWgamepadstate lastControllerState = {0};

static void onButton(CoreButton* map, int n, int key, int action) {
    if(key < 0 || key >= n) {
        return;
    }
    CoreButton b = map[key];
    if(b == 0 || b >= buttonIndex) {
        return;
    }
    if(action == GLFW_RELEASE) {
        buttons[b].upEvents++;
    } else if(action == GLFW_PRESS) {
        buttons[b].downEvents++;
    }
}

/*static void addError(Error& e) {
    const char* description = nullptr;
    int errorCode = glfwGetError(&description);
    if(errorCode == GLFW_NO_ERROR) {
        return;
    }
    e.message.append(": ").append(description);
}

static void handleInputKey(int key, int action) {
    if(action == GLFW_RELEASE) {
        return;
    }
    switch(key) {
        case GLFW_KEY_BACKSPACE:
            if(input.getLength() > inputCursor - 1 && inputCursor > 0) {
                input.remove(inputCursor - 1);
                inputCursor--;
            }
            break;
        case GLFW_KEY_LEFT: inputCursor -= inputCursor > 0; break;
        case GLFW_KEY_RIGHT:
            inputCursor += inputCursor < input.getLength();
            break;
    }
}*/

static void onKey(GLFWwindow*, int key, int scancode, int action, int mods) {
    (void)scancode;
    (void)mods;
    // if(inputActive) {
    //     handleInputKey(key, action);
    // }
    onButton(keyToButton, KEYS, key, action);
}

/*static void addUnicode(uint32 codepoint) {
    if(input.getLength() >= inputLimit) {
        return;
    }
    input.add(codepoint);
    for(int i = input.getLength() - 1; i > inputCursor; i--) {
        Core::swap(input[i], input[i - 1]);
    }
    inputCursor++;
}

static void onChar(GLFWwindow*, uint32 codepoint) {
    if(inputActive) {
        addUnicode(codepoint);
    }
}*/

static void onResize(GLFWwindow*, int width, int height) {
    sizeChanged = true;
    size.data[0] = width;
    size.data[1] = height;
}

static void onMouse(GLFWwindow*, int button, int action, int mods) {
    (void)mods;
    onButton(mouseToButton, MOUSE_BUTTONS, button, action);
}

static void onMouseMove(GLFWwindow*, double x, double y) {
    mousePosition.data[0] = (float)x;
    mousePosition.data[1] = (float)y;
}

bool coreWindowOpen(const CoreWindowOptions* o) {
    if(!glfwInit()) {
        CORE_LOG_ERROR("could not initialize GLFW");
        return true;
    }

    glfwDefaultWindowHints();
    glfwWindowHint(GLFW_VISIBLE, false);
    glfwWindowHint(GLFW_RESIZABLE, true);
    glfwWindowHint(GLFW_DECORATED, !o->fullscreen);
    glfwWindowHint(GLFW_DOUBLEBUFFER, true);

    GLFWmonitor* m = o->fullscreen ? glfwGetPrimaryMonitor() : nullptr;
    window =
        glfwCreateWindow(o->size.data[0], o->size.data[1], o->name, m, nullptr);
    if(window == nullptr) {
        CORE_LOG_ERROR("could not create window");
        coreWindowClose();
        return true;
    }
    size = o->size;
    glfwSetKeyCallback(window, onKey);
    // glfwSetCharCallback(window, onChar);
    glfwSetFramebufferSizeCallback(window, onResize);
    glfwSetMouseButtonCallback(window, onMouse);
    glfwSetCursorPosCallback(window, onMouseMove);
    return false;
}

void coreWindowClose(void) {
    if(window != nullptr) {
        glfwDestroyWindow(window);
        window = nullptr;
    }
    glfwTerminate();
}

void coreWindowShow(void) {
    glfwShowWindow(window);
}

void coreWindowTrapCursor(void) {
    glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
}

void coreWindowFreeCursor(void) {
    glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
}

bool coreWindowIsCursorTrapped(void) {
    return glfwGetInputMode(window, GLFW_CURSOR) == GLFW_CURSOR_DISABLED;
}

const CoreIntVector2* coreWindowGetSize(void) {
    return &size;
}

bool CoreWindowSizeChanged(void) {
    return sizeChanged;
}

bool coreWindowShouldClose(void) {
    return glfwWindowShouldClose(window);
}

void coreWindowRunHandler(CoreWindowRunHandler wr, void* data) {
    runHandlerData = data;
    runHandler = wr == nullptr ? dummyWindowRunHandler : wr;
}

void coreWindowTickHandler(CoreWindowTickHandler t, void* data) {
    tickHandlerData = data;
    tickHandler = t == nullptr ? dummyWindowTickHandler : t;
}

void coreWindowRenderHandler(CoreWindowRenderHandler r, void* data) {
    renderHandlerData = data;
    renderHandler = r == nullptr ? dummyWindowRenderHandler : r;
}

void coreWindowNanosPerTick(i64 nanos) {
    nanosPerTick = nanos <= 0 ? 1 : nanos;
}

static i64 updateClock(Clock* c) {
    i64 nanos = coreGetNanos();
    if(nanos < 0) {
        CORE_LOG_WARNING("Cannot get nanos, using default");
        nanos = 10000000 + c->last;
    }
    i64 diff = nanos - c->last;
    c->last = nanos;
    c->sum -= c->values[c->index];
    c->values[c->index] = diff;
    c->sum += diff;
    c->index = (c->index + 1) % CLOCK_SIZE;
    return diff;
}

static bool searchForGamepad() {
    if(activeController != -1) {
        return true;
    }
    for(int i = GLFW_JOYSTICK_1; i <= GLFW_JOYSTICK_LAST; i++) {
        if(glfwJoystickIsGamepad(i)) {
            activeController = i;
            return true;
        }
    }
    return false;
}

static void checkGamepad() {
    GLFWgamepadstate state;
    if(!glfwGetGamepadState(activeController, &state)) {
        activeController = -1;
        return;
    }
    for(int i = 0; i <= GLFW_GAMEPAD_BUTTON_LAST; i++) {
        CoreButton b = gamepadToButton[i];
        if(b == 0 || b >= buttonIndex) {
            return;
        }
        if(!lastControllerState.buttons[i] && state.buttons[i]) {
            buttons[b].controllerDown++;
        } else if(lastControllerState.buttons[i] && !state.buttons[i]) {
            buttons[b].controllerUp++;
        }
    }
    lastControllerState = state;
}

static void endFrame(void) {
    sizeChanged = false;
    glfwSwapBuffers(window);
    glfwPollEvents();
    if(searchForGamepad()) {
        checkGamepad();
    }
}

static void tickButton(ButtonData* b) {
    bool down = (b->downEvents > 0) || (b->controllerDown > 0);
    bool up = (b->upEvents == b->downEvents) &&
              (b->controllerUp == b->controllerDown);
    if(b->released) {
        b->downTime = 0;
    }
    b->downTime += down;
    b->released = down && up;

    b->downEvents -= b->upEvents;
    b->upEvents = 0;

    b->controllerDown -= b->controllerUp;
    b->controllerUp = 0;
}

static void tick() {
    updateClock(&tps);
    ButtonData* b = buttons;
    for(size_t i = 0; i < buttonIndex; i++) {
        tickButton(b++);
    }
    tickHandler(tickHandlerData);
    lastMousePosition = mousePosition;
}

void coreWindowRun(void) {
    glfwMakeContextCurrent(window);
    tps.last = coreGetNanos();
    fps.last = coreGetNanos();
    i64 lag = 0;
    while(runHandler(runHandlerData)) {
        lag += updateClock(&fps);
        while(lag >= nanosPerTick) {
            lag -= nanosPerTick;
            tick();
        }
        renderHandler(renderHandlerData, (float)lag / (float)nanosPerTick);
        endFrame();
    }
}

float coreWindowTicksPerSecond(void) {
    return (1000000000.0f * CLOCK_SIZE) / (float)tps.sum;
}

float coreWindowFramesPerSecond(void) {
    return (1000000000.0f * CLOCK_SIZE) / (float)fps.sum;
}

/*

void Window::Input::setLimit(int l) {
    inputLimit = l;
    while(input.getLength() > inputLimit) {
        input.removeBySwap(inputLimit);
    }
}

void Window::Input::reset() {
    input.clear();
    inputCursor = 0;
}

void Window::Input::enable() {
    inputActive = true;
}

void Window::Input::disable() {
    inputActive = false;
}

bool Window::Input::isEnabled() {
    return inputActive;
}

static uint32 read(int& index, const char* s) {
    if(s[index] == '\0') {
        return '\0';
    }
    return static_cast<uint32>(s[index++]);
}

static uint32 readUnicode(int& index, const char* s) {
    uint32 c = read(index, s);
    if((c & 0xE0) == 0xC0) {
        c = ((c & 0x1F) << 6) | (read(index, s) & 0x3F);
    } else if((c & 0xF0) == 0xE0) {
        c = ((c & 0xF) << 12) | ((read(index, s) & 0x3F) << 6);
        c |= read(index, s) & 0x3F;
    } else if((c & 0xF8) == 0xF0) {
        c = ((c & 0x7) << 18) | ((read(index, s) & 0x3F) << 12);
        c |= (read(index, s) & 0x3F) << 6;
        c |= read(index, s) & 0x3F;
    }
    return c;
}

void Window::Input::fill(const char* s) {
    int index = 0;
    reset();
    while(true) {
        uint32 c = readUnicode(index, s);
        if(c == '\0') {
            break;
        }
        addUnicode(c);
    }
}

int Window::Input::getCursor() {
    return inputCursor;
}

void Window::Input::setCursor(int index) {
    if(index < 0) {
        inputCursor = 0;
    } else if(index > input.getLength()) {
        inputCursor = input.getLength();
    } else {
        inputCursor = index;
    }
}

const List<uint32>& Window::Input::getUnicode() {
    return input;
}*/

CoreButton coreControlsAdd(const char* name) {
    if(buttonIndex >= BUTTONS) {
        return 0;
    }
    CoreButton b = buttonIndex++;
    snprintf(buttons[b].name, sizeof(buttons[b].name), "%s", name);
    return b;
}

void coreControlsBindKey(CoreButton b, int key) {
    if(key >= 0 && key < KEYS) {
        keyToButton[key] = b;
    }
}

void coreControlsBindGamepad(CoreButton b, int gamepadButton) {
    if(gamepadButton >= 0 && gamepadButton < GAMEPAD_BUTTONS) {
        gamepadToButton[gamepadButton] = b;
    }
}

void coreControlsBindMouse(CoreButton b, int mouseButton) {
    if(mouseButton >= 0 && mouseButton < MOUSE_BUTTONS) {
        mouseToButton[mouseButton] = b;
    }
}

CoreVector2 coreControlsLastMousePosition(void) {
    return lastMousePosition;
}

CoreVector2 coreControlsMousePosition(void) {
    return mousePosition;
}

CoreVector2 coreControlsLeftGamepadAxis(void) {
    return (CoreVector2){{lastControllerState.axes[GLFW_GAMEPAD_AXIS_LEFT_X],
                          lastControllerState.axes[GLFW_GAMEPAD_AXIS_LEFT_Y]}};
}

CoreVector2 coreControlsRightGamepadAxis(void) {
    return (CoreVector2){{lastControllerState.axes[GLFW_GAMEPAD_AXIS_RIGHT_X],
                          lastControllerState.axes[GLFW_GAMEPAD_AXIS_RIGHT_Y]}};
}

float coreControlsLeftGamepadTrigger(void) {
    return lastControllerState.axes[GLFW_GAMEPAD_AXIS_LEFT_TRIGGER];
}

float coreControlsRightGamepadTrigger(void) {
    return lastControllerState.axes[GLFW_GAMEPAD_AXIS_RIGHT_TRIGGER];
}

static const ButtonData* getButton(CoreButton id) {
    return buttons + (id >= buttonIndex ? 0 : id);
}

bool coreControlsDown(CoreButton b) {
    return getButton(b)->downTime > 0;
}

int coreControlsDownTime(CoreButton b) {
    return getButton(b)->downTime;
}

bool coreControlsReleased(CoreButton b) {
    return getButton(b)->released;
}

const char* coreControlsName(CoreButton b) {
    return getButton(b)->name;
}
